import sys
import argparse
import itertools

import tqdm
import map_async
from .preprocess import *
from .unicode import *


def process(args, text):
    text = text.strip()
    if args.sentencize:
        sents = sent_tokenize(text)
    else:
        sents = [text]
    new_sents = []
    for sent in sents:
        if args.normalize:
            sent = normalize(sent)
        delimiter = args.delimiter
        if args.word_tokenize:
            sent = delimiter.join(word_tokenize(sent))
        elif args.morph_tokenize:
            sent = delimiter.join(morph_tokenize(sent, pos=False))
        elif args.pos_tokenize:
            sent = delimiter.join(p for _, p in morph_tokenize(sent, pos=True))
        if args.split_syllables:
            sent = split_syllables(sent)
        elif args.join_jamos:
            sent = join_jamos(sent)
        new_sents.append(sent)
    return args.sentence_delimiter.join(new_sents)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input_path", nargs="?")
    parser.add_argument("output_path", nargs="?")
    parser.add_argument("-P", "--progress", action="store_true", default=False,
                        help="Whether to display the progress bar using tqdm.")
    parser.add_argument("--processes", type=int, default=1,
                        help="Number of processes to utilize.")
    parser.add_argument("-d", "--delimiter", type=str, default=" ",
                        help="The delimiting character to use for splitting "
                             "text into tokens.")
    parser.add_argument("-s", "--sentencize", action="store_true",
                        default=False,
                        help="Split text into sentences, which are delimited "
                             "by sentence delimiters (default is new line).")
    parser.add_argument("--sentence-delimiter", default="\n",
                        help="Character to use for delimiting sentences.")
    parser.add_argument("-n", "--normalize", action="store_true", default=False,
                        help="Whether to perform unicode normalization.")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-w", "--word-tokenize", action="store_true",
                       default=False,
                       help="Further tokenizes tokens into words.")
    group.add_argument("-m", "--morph-tokenize", action="store_true",
                       default=False,
                       help="Further tokenizes tokens into morphemes.")
    group.add_argument("-t", "--pos-tokenize", action="store_true",
                       default=False,
                       help="Further tokenizes tokens into POS tags.")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-p", "--split-syllables", action="store_true",
                       default=False,
                       help="Splits syllables into jamos whenever possible.")
    group.add_argument("-j", "--join-jamos", action="store_true", default=False,
                       help="Joins jamos into syllables whenever possible.")
    args = parser.parse_args()
    in_stream = (open(args.input_path, "r")
                 if args.input_path is not None else sys.stdin)
    out_stream = (open(args.output_path, "w")
                  if args.output_path is not None else sys.stdout)
    desc = "processing text"
    if args.processes == 1:
        it = itertools.starmap(
            process,
            tqdm.tqdm(
                iterable=zip(itertools.repeat(args), in_stream),
                disable=not args.progress,
                desc=desc
            )
        )
    else:
        it = map_async.starmap_async(
            func=process,
            iterable=zip(itertools.repeat(args), in_stream),
            processes=args.processes,
            show_progress=args.progress,
            desc=desc
        )
    for line in it:
        out_stream.write(f"{line}\n")


if __name__ == '__main__':
    main()



__all__ = ["Preprocessor", "normalize", "word_tokenize", "sent_tokenize",
           "morph_tokenize", "sent_word_tokenize", "sent_morph_tokenize"]

import functools

_preprocessor = None


class _Mecab(object):
    def __init__(self, dic_path):
        try:
            import MeCab
        except ImportError:
            raise ImportError("could not import `MeCab`; make sure that "
                              "`mecab-python` is installed by running "
                              "`install_mceab_ko.sh` in the repository. ")
        self._dic_path = dic_path
        self._tagger = MeCab.Tagger("-d {}".format(
            dic_path
        ))

    def parse(self, text):
        nodes = self._tagger.parseToNode(text)

        while nodes:
            form = nodes.surface.strip()
            pos = nodes.feature.split(",")[0]
            nodes = nodes.next

            if pos == "BOS/EOS":
                continue

            yield form, pos


class Preprocessor(object):
    def __init__(self):
        self._mecab = None
        self._twitter = None

    def _init_mecab(self):
        self._mecab = _Mecab("/usr/local/lib/mecab/dic/mecab-ko-dic")

        # Run a test to sacrifice two words
        # There is a bug in mecab that makes it omit first two words.
        try:
            _ = list(self._mecab.parse("mecab mecab"))
            del _
        except UnicodeDecodeError:
            pass

    def _init_twitter(self):
        try:
            import twkorean
        except ImportError:
            raise ImportError("could not import `twkorean`; make sure that "
                              "the package is installed by running "
                              "`install_twkorean.sh` in the repository. ")
        self._twitter = twkorean.TwitterKoreanProcessor()

    def normalize(self, text):
        """Normalize a text using open-korean-text.
        
        Arguments:
            text: text string.
            
        Returns:
            Normalized text string.
        """
        if self._twitter is None:
            self._init_twitter()

        return self._twitter.normalize(text)

    def word_tokenize(self, text):
        """Tokenize a text into space-separated words.
          
        This is the most basic form of tokenization, where we do not wish to
        analyze morphology of each individual word. 
        
        Arguments:
            text: text string.
            
        Returns:
            Generator for a list of space-tokenized words.
        """
        if self._mecab is None:
            self._init_mecab()

        tokens = text.split()
        tokens_it = iter(tokens)

        try:
            token = next(tokens_it)
            index = 0

            for f, pos in self._mecab.parse(text):

                if index >= len(token):
                    if token:
                        yield token

                    token = next(tokens_it)
                    index = 0

                # token_f = token[index:index + len(f)]
                # if token_f != f:
                #     print(token, index, token_f, f)

                if pos.startswith("S"):
                    if index:
                        t, token = token[:index], token[index:]

                        if t:
                            yield t

                        index = 0

                    t, token = token[index:index + len(f)], token[
                                                            index + len(f):]

                    if t:
                        yield t

                    index = 0
                else:
                    index += len(f)

            if token and index:
                yield token[:index]

        except StopIteration:
            pass

    def sent_tokenize(self, text, residual=True):
        """Tokenize a bulk of text into list of sentences (using Mecab-ko).
        
        Arguments:
            text: text string.
            residual: whether to include an incomplete sentence at the end of
                the text.
        Returns:
            Generator that generates a list of sentence strings in their 
            original forms.
        """
        self._init_mecab()

        index = 0

        for f, pos in self._mecab.parse(text):
            index = text.find(f, index)
            index += len(f)

            if pos == "SF":
                sent = text[:index].strip()

                yield sent

                text = text[index:]
                index = 0

        if residual and index > 0:
            sent = text.strip()

            yield sent

    def morph_tokenize(self, text, pos=False):
        """Tokenize a sentence into morpheme tokens (using Mecab-ko).
        
        Arguments:
            text: sentence string.
            pos: whether to include part-of-speech tags.
            
        Returns:
            If pos is False, then a generator of morphemes is returned. 
            Otherwise, a generator of morpheme and pos tuples is returned.
        """
        self._init_mecab()

        if pos:
            for item in self._mecab.parse(text):
                yield item
        else:
            for f, _ in self._mecab.parse(text):
                yield f

    def sent_morph_tokenize(self, text, residual=True, pos=False):
        """Tokenize a bulk of text into list of sentences (using Mecab-ko).

        Each sentence is a list of morphemes. This is slightly more efficient than
        tokenizing text into sents and morphemes in succession.

        Arguments:
            text: text string.
            residual: whether to include an incomplete sentence at the end of
                the text.
            pos: whether to include part-of-speech tag.
        Returns:
            If pos is False, then a generator of morphemes list is returned. 
            Otherwise, a generator of morpheme and pos tuples list is returned.
        """
        self._init_mecab()

        sent = []

        for f, p in self._mecab.parse(text):
            if pos:
                sent.append((f, p))
            else:
                sent.append(f)

            if p == "SF":
                yield sent
                sent = []

        if residual and sent:
            yield sent

    def sent_word_tokenize(self, text, residual=True):
        """Tokenize a bulk of text into list of sentences (using Mecab-ko).

        Each sentence is a list of words. This is slightly more efficient than
        tokenizing text into sents and words in succession.

        Arguments:
            text: text string.
            residual: whether to include an incomplete sentence at the end of
                the text.
        Returns:
            A generator of words list. 
        """
        self._init_mecab()

        sent = []
        tokens = text.split()
        tokens_it = iter(tokens)

        try:
            token = next(tokens_it)
            index = 0
            yield_sent = False

            for f, pos in self._mecab.parse(text):

                if index >= len(token):
                    if token:
                        sent.append(token)

                    token = next(tokens_it)
                    index = 0

                    if yield_sent:
                        yield sent
                        yield_sent = False
                        sent = []

                # assert token[index:index + len(f)] == f

                if pos.startswith("S"):
                    if index:
                        t, token = token[:index], token[index:]

                        if t:
                            sent.append(t)

                        index = 0

                    t, token = token[index:index + len(f)], token[
                                                            index + len(f):]

                    if t:
                        sent.append(t)

                    index = 0

                    if pos == "SF":
                        yield_sent = True
                else:
                    index += len(f)

            if token and index:
                sent.append(token[:index])

            if sent and (yield_sent or residual):
                yield sent

        except StopIteration:
            pass


def normalize(text, *args, **kwargs):
    global _preprocessor

    if _preprocessor is None:
        _preprocessor = Preprocessor()

    return _preprocessor.normalize(text, *args, **kwargs)


def morph_tokenize(text, *args, **kwargs):
    global _preprocessor

    if _preprocessor is None:
        _preprocessor = Preprocessor()

    return _preprocessor.morph_tokenize(text, *args, **kwargs)


def sent_tokenize(text, *args, **kwargs):
    global _preprocessor

    if _preprocessor is None:
        _preprocessor = Preprocessor()

    return _preprocessor.sent_tokenize(text, *args, **kwargs)


def word_tokenize(text, *args, **kwargs):
    global _preprocessor

    if _preprocessor is None:
        _preprocessor = Preprocessor()

    return _preprocessor.word_tokenize(text, *args, **kwargs)


def sent_word_tokenize(text, *args, **kwargs):
    global _preprocessor

    if _preprocessor is None:
        _preprocessor = Preprocessor()

    return _preprocessor.sent_word_tokenize(text, *args, **kwargs)


def sent_morph_tokenize(text, *args, **kwargs):
    global _preprocessor

    if _preprocessor is None:
        _preprocessor = Preprocessor()

    return _preprocessor.sent_morph_tokenize(text, *args, **kwargs)


functools.update_wrapper(normalize, Preprocessor.normalize)
functools.update_wrapper(sent_tokenize, Preprocessor.sent_tokenize)
functools.update_wrapper(morph_tokenize, Preprocessor.morph_tokenize)
functools.update_wrapper(word_tokenize, Preprocessor.word_tokenize)
functools.update_wrapper(sent_word_tokenize, Preprocessor.sent_word_tokenize)
functools.update_wrapper(sent_morph_tokenize, Preprocessor.sent_morph_tokenize)



from setuptools import setup

setup(
    name='hangul-utils',
    version="0.4.1",
    license="GPL",
    description='An integrated library for Korean preprocessing.',
    author='Kang Min Yoo',
    author_email='kaniblurous@gmail.com',
    url='https://github.com/kaniblu/hangul-utils',
    packages=['hangul_utils'],
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Natural Language :: Korean",
        "Programming Language :: Java",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Linguistic"
    ],
    platforms=[
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
    ],
    keywords=[
        "hangul-utils",
        "morphological analyzer",
        "morphology",
        "analyzer",
        "korean",
        "tokenizer",
        "sentence tokenizer",
        "word tokenizer",
        "pos tagger",
        "natural language processing",
        "jamo",
        "korean character",
    ],
    install_requires=[
        "tqdm",
        "six",
        "jpype1;python_version<='2.7'",
        "jpype1-py3;python_version>='3.5'",
        "mecab-python==0.996-ko-0.9.2",
        "map-async==1.1"
    ],
    dependency_links=[
        "git+https://bitbucket.org/eunjeon/"
        "mecab-python-0.996#egg=mecab-python-0.996-ko-0.9.2"
    ],
    entry_points={
        "console_scripts": [
            "hangul-process = hangul_utils.run:main"
        ]
    }
)
